%{
/*MA****************************************************************/
/*                                                                 */
/*  FILE     tmod.y                                                */
/*  MODULE   tmod                                                  */
/*  PROGRAM  TMod                                                  */
/*  AUTHOR   Helmut Schmid, IMS, University of Stuttgart           */
/*                                                                 */
/*ME****************************************************************/

#include <assert.h>
#include <stdio.h>

#include <iostream>
using std::cerr;

#include "tree.h"

extern int  yylineno;
extern char *yytext;

void yyerror(char *text);
int yylex( void );

char *HeadRulesFileName=NULL;
char *ScoreFileName=NULL;
char *GovFileName=NULL;
%}

%union {
  Tree   *tree;
  Trees  *trees;
  char   *name;
}

%token <name>      NAME IGNORE
%type  <tree>      TREE
%type  <trees>     TREES

%%

ALL:        ALL TREE                { $2->process(); delete $2; }
          | ALL IGNORE              { printf("%s\n",$2); free($2); }
          | {}
          ;

TREE:       '(' NAME TREES ')'      { $$=new Tree($2,$3); }
          | NAME                    { $$=new Tree($1); }
          | '(' NAME ')'            { $$=new Tree($2); }
          ;

TREES:      TREES TREE              { $1->push_back($2); $$=$1; }
          | TREE                    { $$ = new Trees; $$->push_back($1); }
          ;

%%


extern FILE  *yyin;

/*FA****************************************************************/
/*                                                                 */
/*  yyerror                                                        */
/*                                                                 */
/*FE****************************************************************/

void yyerror(char *text)

{
  cerr << yylineno << ": " << text << " at: ";
  cerr << yytext << "\naborted.\n";
  exit(1);
}


/*FA****************************************************************/
/*                                                                 */
/*  usage                                                          */
/*                                                                 */
/*FE****************************************************************/

void usage()

{
  fprintf(stderr,"\nUsage: tmod [options] infile\n");
  fprintf(stderr,"\nOPTIONS:\n");
  fprintf(stderr,"-r\textract treebank grammar and lexicon\n");
  fprintf(stderr,"-t\tprint trace nodes\n");
  fprintf(stderr,"-new do not collapse epsilon expansions into a single complex symbol and ignore \"=\"-coreferences\n");
  fprintf(stderr,"-g <file>\textract head governor tuples to <file>\n");
  fprintf(stderr,"-H <file>\tmark the heads of the extracted rules\n");
  fprintf(stderr,"-m <file>\tread symbol mapping from <file>\n");
  fprintf(stderr,"-pm <file>\tread preposition mapping from <file>\n");
  fprintf(stderr,"-u \tundo markovization\n");
  fprintf(stderr,"-c\tconvert slash features to coreference tags\n");
  fprintf(stderr,"-e \tprint empty categories\n");
  fprintf(stderr,"-C \tprint constituents\n");
  fprintf(stderr,"-n\tprint a list of constituents rather than a parse tree\n");
  fprintf(stderr,"-R\tprint the individual rules with heads\n");
  fprintf(stderr,"-l\tprint the product of the lexical scores for each parse\n");
  fprintf(stderr,"-p\tpretty print mode\n");
  fprintf(stderr,"-q\tquiet mode\n");
  fprintf(stderr,"\n");
  exit(0);
}


/*FA****************************************************************/
/*                                                                 */
/*  get_flags                                                      */
/*                                                                 */
/*FE****************************************************************/

void get_flags( int *argc, char **argv )

{
  for( int i=1; i<*argc; i++ ) {
    if (strcmp(argv[i],"-q") == 0) {
      Verbose = false;
      argv[i] = NULL;
    }
    else if (strcmp(argv[i],"-r") == 0) {
      PrintGrammar = true;
      argv[i] = NULL;
    }
    else if (strcmp(argv[i],"-r") == 0) {
      PrintGrammar = true;
      argv[i] = NULL;
    }
    else if (strcmp(argv[i],"-R") == 0) {
      PrintRules = true;
      argv[i] = NULL;
    }
    else if (strcmp(argv[i],"-p") == 0) {
      PrettyPrint = true;
      argv[i] = NULL;
    }
    else if (strcmp(argv[i],"-n") == 0) {
      PrintNodes = true;
      argv[i] = NULL;
    }
    else if (strcmp(argv[i],"-t") == 0) {
      PrintTraces = true;
      argv[i] = NULL;
    }
    else if (strcmp(argv[i],"-c") == 0) {
      ConvertTraces = true;
      argv[i] = NULL;
    }
    else if (strcmp(argv[i],"-u") == 0) {
      UndoMarkov = true;
      argv[i] = NULL;
    }
    else if (strcmp(argv[i],"-C") == 0) {
      PrintConstituents = true;
      argv[i] = NULL;
    }
    else if (strcmp(argv[i],"-e") == 0) {
      EmptyCategories = true;
      argv[i] = NULL;
    }
    else if (strcmp(argv[i],"-new") == 0) {
      NewGrammar = true;
      argv[i] = NULL;
    }
    else if (strcmp(argv[i],"-l") == 0) {
      ScoreFileName = argv[i+1];
      argv[i] = NULL;
      argv[++i] = NULL;
      LexScore = true;
    }
    else if (strcmp(argv[i],"-g") == 0 && i < *argc+1) {
      GovFileName = argv[i+1];
      argv[i] = NULL;
      argv[++i] = NULL;
    }
    else if (strcmp(argv[i],"-m") == 0 && i < *argc+1) {
      MapFileName = argv[i+1];
      argv[i] = NULL;
      argv[++i] = NULL;
    }
    else if (strcmp(argv[i],"-H") == 0 && i < *argc+1) {
      HeadRulesFileName = argv[i+1];
      WithHeads = true;
      argv[i] = NULL;
      argv[++i] = NULL;
    }
    else if (strcmp(argv[i],"-pm") == 0 && i < *argc+1) {
      PrepFileName = argv[i+1];
      argv[i] = NULL;
      argv[++i] = NULL;
    }
    else if (strcmp(argv[i],"-h") == 0)
      usage();
  }
  if (LexScore && !WithHeads) {
    fprintf(stderr,"\nError: The lexscore option -l requires the specification of a head rules file via the option -H\n\n");
    exit(1);
  }

  // remove flags from the argument list
  int k;
  for( int i=k=1; i<*argc; i++)
    if (argv[i] != NULL)
      argv[k++] = argv[i];
  *argc = k;
}


/*FA****************************************************************/
/*                                                                 */
/*  main                                                           */
/*                                                                 */
/*FE****************************************************************/

int main( int argc, char *argv[] )

{
  FILE *file;

  get_flags(&argc, argv);

  if (MapFileName)
    read_mapping(MapFileName, CatMap);
  if (WithHeads)
    read_head_rules(HeadRulesFileName);
  if (LexScore)
    read_lexscores(ScoreFileName);
  if (PrepFileName)
    read_mapping(PrepFileName, PrepMap);

  govfile = NULL;
  if (GovFileName) {
    if (!WithHeads)
      fprintf(stderr,"Warning: Option -g ignored without option -H!\n");
    else if ((govfile = fopen(GovFileName,"wt")) == 0) {
      fprintf(stderr,"\nError: Cannot open output file \"%s\"\n\n",
	      GovFileName);
      exit(1);
    }
  }

  if (argc > 1) {
    if ((file = fopen(argv[1],"rt")) == NULL) {
      fprintf(stderr,"\nError: Cannot open input file \"%s\"\n\n", argv[1]);
      exit(1);
    }
  }
  else
    file = stdin;

  if (argc > 2) {
    if ((outfile = fopen(argv[2],"wt")) == NULL) {
      fprintf(stderr,"\nError: Cannot output output file \"%s\"\n\n",argv[2]);
      exit(1);
    }
  }
  else
    outfile = stdout;

  if (PrintGrammar) {
    if (argc > 3) {
      if ((lexfile = fopen(argv[3],"wt")) == NULL) {
	fprintf(stderr,"\nError: Cannot open input file \"%s\"\n\n", argv[3]);
	exit(1);
      }
    }
    else
      lexfile = outfile;
  }

  yyin = file;
  try {
    yyparse();
    fclose(file);
  }
  catch(const char* p) {
    cerr << "\n" << p << "\n\n";
    exit(1);
  }

  if (PrintGrammar) {
    if (Verbose)
      fputs("\nPrinting the grammar rules ...", stderr);
    print_grammar();
  }

  if (Verbose)
    fputs(" finished\n", stderr);
}
