
/*MA****************************************************************/
/*                                                                 */
/*     File: tree2.C                                               */
/*   Author: Helmut Schmid                                         */
/*  Purpose:                                                       */
/* Modified: Mon Jun 30 11:46:14 2008 (schmid)                     */
/*                                                                 */
/*ME****************************************************************/

#include <stdio.h>

#include <iostream>
using std::cerr;


#include "tree.h"

int Trace_Number;


/*FA****************************************************************/
/*                                                                 */
/*  Tree::slash_cat                                                */
/*                                                                 */
/*  Task: Returns a pointer to the last slash feature (if present) */
/*                                                                 */
/*FE****************************************************************/

char *Tree::slash_cat( char *d )

{
  char *p=last_char(label);
  if (*p == '') {
    int depth=1;
    for( p--; p>label+2; p-- )
      if (*p == '')
	depth++;
      else if (*p == '')
	if (--depth == 0 && *(p-1) == '#' &&
	    (*(p-2) == '+' || *(p-2) == '%'))
	  {
	    *d = *(p-2);
	    return p;
	  }
  }

  return NULL;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::find_trace                                               */
/*                                                                 */
/*FE****************************************************************/

char *Tree::find_trace( char *s, char d, bool &terminal )

{
  char *le=last_char(label);
  char *se=last_char(s);

  // find matching slash feature
  for( ; le > label && se >= s && *le == *se; le--, se--)
    if (se == s) {
      // matching slash feature recognized
      le--;
      if (le > label && *le == '#') {
	terminal = true;
	le--;
      }
      else
	terminal = false;
      if (*le == d)
	return le;
    }
  
  return NULL;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_trace_feature                                        */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_trace_feature( char c )

{
  char buffer[100];
  sprintf(buffer,"%d",Trace_Number);
  add_markup(buffer,c);
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::delete_slashes_down                                      */
/*                                                                 */
/*FE****************************************************************/

bool Tree::delete_slashes_down( char *sc, char d )

{
  bool term;
  char *trace_start = find_trace( sc, d, term );

  if (trace_start == NULL)
    return false;

  if (term) {
    *trace_start = 0;
    if (d == '_')
      add_trace_feature('-');
    else
      add_trace_feature('=');
    return true;
  }

  for( size_t i=0; i<daughters.size(); i++ )
    if (daughters[i]->delete_slashes_down( sc, d )) {
      *trace_start = 0;
      return true;
    }

  return false;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::delete_slashes_up                                        */
/*                                                                 */
/*FE****************************************************************/

bool Tree::delete_slashes_up( char *sc, char d )

{
  bool term;
  char *trace_start = find_trace( sc, d, term );

  if (trace_start && mother->delete_slashes_up(sc,d)) {
    *trace_start = 0;
    return true;
  }

  if (d == '_' || d == '+')
    d = '_';
  else
    d = '=';
  for( size_t i=0; i<daughters.size(); i++ )
    if (daughters[i]->delete_slashes_down( sc, d ))
      return true;
  
  return false;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::convert                                                  */
/*                                                                 */
/*FE****************************************************************/

bool Tree::convert( bool &remaining_traces )

{
  bool trace_found=false;
  char d;
  char *p;

  // Check for +#.*
  while ((p=slash_cat(&d)) != NULL) {
    Trace_Number++;
    char *sf = strdup(p);
    bool successful = delete_slashes_up(sf,d);
    free(sf);
    if (!successful) {
      remaining_traces = true;
      Trace_Number--;
      break;
    }
    if (*(p-2)) // slash feature not deleted yet
      *(p-2) = 0;
    trace_found = true;
    add_trace_feature('-');
  }

  for( size_t i=0; i<daughters.size(); i++ )
    trace_found = daughters[i]->convert(remaining_traces) || trace_found;

  return trace_found;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::delete_slash_features                                    */
/*                                                                 */
/*FE****************************************************************/

void Tree::delete_slash_features()

{
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->delete_slash_features();

  for( char *p=label; *p; p++ )
    if (*p == '_' || *p == '+' || *p == '%' || *p == '=') {
      char *q = p+1;
      if (*q == '#')
	q++;
      if (*q == '')
	*p = 0; // remove tag
      return;
    }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::print_ec2                                                */
/*                                                                 */
/*FE****************************************************************/

int Tree::print_ec2( int pos )

{
  // Special treatment of (SBAR (-NONE- 0)(S (-NONE *)))
  if (matching_base_cat("SBAR") && daughters.size() == 2 &&
      daughters[0]->matching_base_cat("-NONE-") &&
      daughters[1]->label[0] == 'S' &&
      daughters[1]->daughters[0]->matching_base_cat("-NONE-"))
    {
      fprintf(outfile,"\t%d %s SBARXS\n", pos, 
	      daughters[1]->daughters[0]->daughters[0]->label);
      return pos;
    }

  bool index_found=false;
  char *p;
  for( p=label; *p; p++ ); // find the end of the string
  for( p--; p > label; p-- ) { // find the index
    if (*p >= '0' && *p <= '9')
      continue;
    if (*p == '-' && *(p+1) != 0) {
      if (daughters.size() == 0 && mother->matching_category("-NONE-"))
	fprintf(outfile,"\t%d %s %s\n", pos, label, mother->mother->label);
      else if (daughters.size() > 0)
	fprintf(outfile,"\t%d %s\n", pos, label);
      index_found = true;
    }
    break;
  }
  if (!index_found && mother && mother->matching_category("-NONE-"))
    fprintf(outfile,"\t%d %s %s\n", pos, label, mother->mother->label);

  if (daughters.size() == 0) {
    if (!mother->matching_category("-NONE-"))
      pos++;
  }
  else
    for( size_t i=0; i<daughters.size(); i++ )
      pos = daughters[i]->print_ec2(pos);

  return pos;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::print_ec                                                 */
/*                                                                 */
/*FE****************************************************************/

void Tree::print_ec()

{
  static int N=0;

  fprintf(outfile,"%d\n", ++N);
  print_ec2(0);
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::print_constituents                                       */
/*                                                                 */
/*FE****************************************************************/

void Tree::print_constituents( int &pos )

{
  if (matching_category("-NONE-"))
    return;

  if (daughters.size() == 1 && daughters[0]->daughters.size() == 0) {
    pos++;
    return;
  }

  int start = pos;
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->print_constituents( pos );

  if (mother && start < pos)
    printf("%s %d %d\n", label, start, pos);
}
